#ifndef STREAM_HPP_INCLUDED
#define STREAM_HPP_INCLUDED


#include<cstdio>
#include<cstddef>
#include<cstdint>
#include<cstdlib>
#include<cstring>
#include<string>
#include<memory>
#include<utility>




using  vlint_t =  int32_t;
using uvlint_t = uint32_t;


/*
1byte ,  7bits 1xxxxxxx

2bytes, 12bits 0LLLxxxx xxxxxxxx
3bytes, 20bits 0LLLxxxx xxxxxxxx xxxxxxxx
4bytes, 28bits 0LLLxxxx xxxxxxxx xxxxxxxx xxxxxxxx
5bytes, 36bits 0LLLxxxx xxxxxxxx xxxxxxxx xxxxxxxx xxxxxxxx


Lの部分3ビットを取り出して2を足した値が、可変長整数のバイト数
*/


inline
size_t
get_vlint_size(uvlint_t  i)
{
  size_t  size;

       if(i <= 0x0000007F){size = 1;}
  else if(i <= 0x00000FFF){size = 2;}
  else if(i <= 0x000FFFFF){size = 3;}
  else if(i <= 0x0FFFFFFF){size = 4;}
  else                    {size = 5;}

  return size;
}


class
Stream
{
public:
  virtual ~Stream(){};

  virtual size_t  get_pos(           ) const = 0;
  virtual void    set_pos(size_t  pos)       = 0;

  virtual bool  test_end()   const = 0;
  virtual bool  test_error() const = 0;

  virtual uint8_t  get(          ) = 0;
  virtual void     put(uint8_t  c) = 0;

  uint16_t  get_be16()
  {
    uint16_t  u = 0;

    u |= (get()<<8);
    u |= (get()   );

    return u;
  }


  uint32_t  get_be32()
  {
    uint32_t  u = 0;

    u |= (get()<<24);
    u |= (get()<<16);
    u |= (get()<< 8);
    u |= (get()    );

    return u;
  }


  vlint_t   get_vlint()
  {
    auto  i = get_uvlint();

    const bool  negative = (i&1);

    i >>= 1;

    return(negative? -i:i);
  }


  uvlint_t  get_uvlint()
  {
    uint8_t  first_byte = get();

      if(first_byte&0x80)
      {
        return(first_byte&0x7F);
      }


    size_t  size = (first_byte>>4);

    uint64_t  i;

      switch(size)
      {
        case(0):
          i = ((first_byte&0x0F)<<8);
          i |= get();
          break;
        case(1):
          i = ((first_byte&0x0F)<<16);
          i |= (get()<<8);
          i |= (get()   );
          break;
        case(2):
          i = ((first_byte&0x0F)<<24);
          i |= (get()<<16);
          i |= (get()<< 8);
          i |= (get()    );
          break;
        case(3):
          i  = (get()<<24);
          i |= (get()<<16);
          i |= (get()<< 8);
          i |= (get()    );
          break;
      }


    return i;
  }


  uint16_t  put_be16(uint16_t  c)
  {
    put((c>>8)&0xFF);
    put((c   )&0xFF);

    return c;
  }


  uint32_t  put_be32(uint32_t  c)
  {
    put((c>>24)&0xFF);
    put((c>>16)&0xFF);
    put((c>> 8)&0xFF);
    put((c    )&0xFF);

    return c;
  }

  void   put_vlint(vlint_t  i)
  {
    const bool  negative = (i < 0);

    i = std::abs(i)<<1;

      if(negative)
      {
        i |= 1;
      }


    put_uvlint(i);
  }


  void  put_uvlint(uvlint_t  i)
  {
      if(i <= 0x7F)
      {	
        put(i|0x80);

        return;
      }


      switch(get_vlint_size(i))
      {
        case(2):
          put(i>>8);
          put(i&0xFF);
          break;
        case(3):
          put((1<<4)|(i>>16));
          put((i>>8)&0xFF);
          put((i>>0)&0xFF);
          break;
        case(4):
          put((2<<4)|(i>>24));
          put((i>>16)&0xFF);
          put((i>> 8)&0xFF);
          put((i>> 0)&0xFF);
          break;
        case(5):
          put(3<<4);
          put((i>>24)&0xFF);
          put((i>>16)&0xFF);
          put((i>> 8)&0xFF);
          put((i>> 0)&0xFF);
          break;
      }
  }


  void  write(const void*  buf, size_t  bytes)
  {
    auto  src = static_cast<const char*>(buf);

      while(bytes--)
      {
        put(*src++);
      }
  }


  void  write(const std::string&  s)
  {
      for(auto  c: s)
      {
        put(c);
      }


    put(0);
  }


  void  write(const std::u16string&  s)
  {
      for(auto  c: s)
      {
        put_be16(c);
      }


    put_be16(0);
  }


  void   read(void*  buf, size_t  bytes)
  {
    auto  dst = static_cast<char*>(buf);

      while(bytes--)
      {
        *dst++ = get();
      }
  }


  void   read(std::string&  s)
  {
      for(;;)
      {
        auto  c = get();

          if(!c || test_end())
          {
            break;
          }


        s.push_back(c);
      }
  }


  void   read(std::u16string&  s)
  {
      for(;;)
      {
        auto  c = get_be16();

          if(!c || test_end())
          {
            break;
          }


        s.push_back(c);
      }
  }

};




class
MemoryStream: public Stream
{
  std::unique_ptr<std::string>  unique_s_;

  std::string*  s_;

  size_t  offset_;

public:
  MemoryStream():
  unique_s_(new std::string),
  s_(unique_s_.get()),
  offset_(0)
  {
  }

  MemoryStream(std::string&  s):
  s_(&s),
  offset_(0)
  {
  }

  const std::string&  get_buffer() const
  {
    return *s_;
  }

  size_t  get_pos() const
  {
    return offset_;
  }


  void    set_pos(size_t  pos)
  {
    offset_ = pos;
  }


  bool  test_end() const
  {
    return (offset_ >= s_->size());
  }

  bool  test_error() const
  {
    return 0;
  }

  uint8_t  get()
  {
    return static_cast<uint8_t>((*s_)[offset_++]);
  }

  void  put(uint8_t  c)
  {
      if(offset_ < s_->size())
      {
        (*s_)[offset_] = c;
      }

    else
      {
        s_->push_back(c);
      }


    offset_ += 1;
  }

};




class
FileStream: public Stream
{
  FILE*  f_;

public:
  FileStream(FILE*  f): f_(f){}
 ~FileStream()
  {
      if(f_)
      {
        fclose(f_);
      }
  }

  size_t  get_pos() const
  {
    return ftell(f_);
  }


  void  set_pos(size_t  pos)
  {
    fseek(f_,pos,SEEK_SET);
  }

  bool  test_end() const
  {
    return feof(f_);
  }

  bool  test_error() const
  {
    return ferror(f_);
  }

  uint8_t  get()
  {
    return fgetc(f_);
  }

  void  put(uint8_t  c)
  {
    fputc(c,f_);
  }

};




#endif




