#ifndef PIXELMAP_HPP_INCLUDED
#define PIXELMAP_HPP_INCLUDED


#include<cstdint>
#include<cstring>
#include<vector>


template<typename  PIXEL>
class
PixelIterator
{
  uint8_t*  pointer_;
  const int   pitch_;

public:
  PixelIterator(PIXEL*  pointer, int  pitch):
  pointer_(reinterpret_cast<uint8_t*>(pointer)),
  pitch_(pitch)
  {
  }


  PIXEL*  get()
  {
    PIXEL*  p = reinterpret_cast<PIXEL*>(pointer_)         ;
                                         pointer_ += pitch_;

    return p;
  }

};


template<typename  PIXEL>
class
ConstPixelIterator
{
  const uint8_t*  pointer_;
  const int   pitch_;

public:
  ConstPixelIterator(const PIXEL*  pointer, int  pitch):
  pointer_(reinterpret_cast<const uint8_t*>(pointer)),
  pitch_(pitch)
  {
  }


  const PIXEL*  get()
  {
    const PIXEL*  p = reinterpret_cast<const PIXEL*>(pointer_)         ;
                                                     pointer_ += pitch_;

    return p;
  }

};


template<typename  PIXEL>
class
PixelMap
{
protected:
  std::vector<PIXEL>  buffer_;

  int   width_;
  int  height_;
  int   pitch_;

  uint8_t*  pointer_;

public:
  PixelMap()
  {
  }


  PixelMap(int  w, int  h)
  {
    resize(w,h);
  }


  PixelMap(PIXEL*  pointer, int  w, int  h, int  pitch)
  {
    width_  = w;
    height_ = h;
    pitch_  = pitch;

    pointer_ = reinterpret_cast<uint8_t*>(pointer);
  }


  PixelMap(PixelMap<PIXEL>&&  rhs)
  {
    *this = std::move(rhs);
  }


  PixelMap(const PixelMap<PIXEL>&  rhs)
  {
    *this = rhs;
  }


  PixelMap<PIXEL>&  operator=(PixelMap<PIXEL>&&  rhs)
  {
    width_  = rhs.width_;
    height_ = rhs.height_;
    pitch_  = rhs.pitch_;

    buffer_ = std::move(rhs.buffer_);

    pointer_ = rhs.pointer_;

    return *this;
  }


  PixelMap<PIXEL>&  operator=(const PixelMap<PIXEL>&  rhs)
  {
    width_  = rhs.width_;
    height_ = rhs.height_;
    pitch_  = rhs.pitch_;

    buffer_ = rhs.buffer_;

    pointer_ = reinterpret_cast<uint8_t*>(buffer_.data());

    return *this;
  }


  int  get_width()  const{return width_ ;}
  int  get_height() const{return height_;}

  PIXEL&  get_pixel(int  x, int  y)
  {
    auto  p = pointer_;

    p += (pitch_       *y);
    p += (sizeof(PIXEL)*x);

    return *reinterpret_cast<PIXEL*>(p);
  }


  const PIXEL&  get_pixel(int  x, int  y) const
  {
    auto  p = pointer_;

    p += (pitch_       *y);
    p += (sizeof(PIXEL)*x);

    return *reinterpret_cast<PIXEL*>(p);
  }

  void  fill()
  {
    std::memset(pointer_,0,(pitch_*height_));
  }

  void set_width(int  w)
  {
    width_  = w;
    pitch_  = sizeof(PIXEL)*w;
  }

  void set_height(int  h)
  {
    height_ = h;
  }

  void resize()
  {
    buffer_.resize(pitch_*height_);

    pointer_ = reinterpret_cast<uint8_t*>(buffer_.data());
  }

  void resize(int  w, int  h)
  {
    width_  = w;
    height_ = h;
    pitch_  = sizeof(PIXEL)*w;

    buffer_.resize(pitch_*height_);

    pointer_ = reinterpret_cast<uint8_t*>(buffer_.data());
  }


  PixelIterator<PIXEL>
  get_iterator(int  x, int  y)
  {
    return PixelIterator<PIXEL>(&get_pixel(x,y),pitch_);
  }


  ConstPixelIterator<PIXEL>
  get_const_iterator(int  x, int  y) const
  {
    return ConstPixelIterator<PIXEL>(&get_pixel(x,y),pitch_);
  }


  void  draw_point(PIXEL  pixel, int  x, int  y)
  {
    get_pixel(x,y) = pixel;
  }

  void  draw_vline(PIXEL  pixel, int  x, int  y, int  length)
  {
      while(length--)
      {
        draw_point(pixel,x,y++);
      }
  }

  void  draw_hline(PIXEL  pixel, int  x, int  y, int  length)
  {
      while(length--)
      {
        draw_point(pixel,x++,y);
      }
  }

};


using PixelMap16 = PixelMap<uint16_t>;
using PixelMap32 = PixelMap<uint32_t>;


void  transfer(const PixelMap16&  src, int  src_x, int  src_y, int  src_w, int  src_h,
                     PixelMap16&  dst, int  dst_x, int  dst_y);

void  transform(const PixelMap16&  src,
                      PixelMap16&  dst);

void  transfer(const PixelMap16&  src, int  src_x, int  src_y, int  src_w, int  src_h,
                     PixelMap32&  dst, int  dst_x, int  dst_y);




#endif




